#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from functools import cached_property

from skale_contracts.projects.ima import SchainImaContract
from skale_contracts.project_factory import SkaleProject

from skale.skale_base import SkaleBase
from skale.contracts.ima.schain.message_proxy_for_schain import MessageProxyForSchain
from skale.contracts.ima.schain.token_manager_linker import TokenManagerLinker
from skale.contracts.ima.schain.token_manager_eth import TokenManagerETH
from skale.contracts.ima.schain.token_manager_erc20 import TokenManagerERC20
from skale.contracts.ima.schain.token_manager_erc721 import TokenManagerERC721
from skale.contracts.ima.schain.token_manager_erc721_wmt import (
    TokenManagerERC721WithMetadata,
)
from skale.contracts.ima.schain.token_manager_erc1155 import TokenManagerERC1155
from skale.contracts.ima.schain.community_locker import CommunityLocker


class SchainIma(SkaleBase):
    @property
    def project_name(self) -> SkaleProject:
        return SkaleProject.SCHAIN_IMA

    @cached_property
    def message_proxy_for_schain(self) -> MessageProxyForSchain:
        return MessageProxyForSchain(self, SchainImaContract.MESSAGE_PROXY_FOR_SCHAIN)

    @cached_property
    def community_locker(self) -> CommunityLocker:
        return CommunityLocker(self, SchainImaContract.COMMUNITY_LOCKER)

    @cached_property
    def token_manager_eth(self) -> TokenManagerETH:
        return TokenManagerETH(self, SchainImaContract.TOKEN_MANAGER_ETH)

    @cached_property
    def token_manager_erc20(self) -> TokenManagerERC20:
        return TokenManagerERC20(self, SchainImaContract.TOKEN_MANAGER_ERC20)

    @cached_property
    def token_manager_erc721(self) -> TokenManagerERC721:
        return TokenManagerERC721(self, SchainImaContract.TOKEN_MANAGER_ERC721)

    @cached_property
    def token_manager_erc721_wmt(self) -> TokenManagerERC721WithMetadata:
        return TokenManagerERC721WithMetadata(
            self, SchainImaContract.TOKEN_MANAGER_ERC721_WITH_META
        )

    @cached_property
    def token_manager_erc1155(self) -> TokenManagerERC1155:
        return TokenManagerERC1155(self, SchainImaContract.TOKEN_MANAGER_ERC1155)

    @cached_property
    def token_manager_linker(self) -> TokenManagerLinker:
        return TokenManagerLinker(self, SchainImaContract.TOKEN_MANAGER_LINKER)


def spawn_skale_ima_lib(schain_ima: SchainIma) -> SchainIma:
    """Clone skale ima object with the same wallet"""
    return SchainIma(schain_ima._endpoint, schain_ima.instance.address, schain_ima.wallet)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from functools import cached_property

from skale_contracts.projects.ima import MainnetImaContract
from skale_contracts.project_factory import SkaleProject

from skale.skale_base import SkaleBase
from skale.contracts.ima.mainnet.message_proxy_for_mainnet import MessageProxyForMainnet
from skale.contracts.ima.mainnet.linker import Linker
from skale.contracts.ima.mainnet.community_pool import CommunityPool
from skale.contracts.ima.mainnet.deposit_box_eth import DepositBoxEth
from skale.contracts.ima.mainnet.deposit_box_erc20 import DepositBoxERC20
from skale.contracts.ima.mainnet.deposit_box_erc721 import DepositBoxERC721
from skale.contracts.ima.mainnet.deposit_box_erc721_wmt import (
    DepositBoxERC721WithMetadata,
)
from skale.contracts.ima.mainnet.deposit_box_erc1155 import DepositBoxERC1155


class SkaleIma(SkaleBase):
    @property
    def project_name(self) -> SkaleProject:
        return SkaleProject.MAINNET_IMA

    @cached_property
    def message_proxy_for_mainnet(self) -> MessageProxyForMainnet:
        return MessageProxyForMainnet(self, MainnetImaContract.MESSAGE_PROXY_FOR_MAINNET)

    @cached_property
    def linker(self) -> Linker:
        return Linker(self, MainnetImaContract.LINKER)

    @cached_property
    def community_pool(self) -> CommunityPool:
        return CommunityPool(self, MainnetImaContract.COMMUNITY_POOL)

    @cached_property
    def deposit_box_eth(self) -> DepositBoxEth:
        return DepositBoxEth(self, MainnetImaContract.DEPOSIT_BOX_ETH)

    @cached_property
    def deposit_box_erc20(self) -> DepositBoxERC20:
        return DepositBoxERC20(self, MainnetImaContract.DEPOSIT_BOX_ERC20)

    @cached_property
    def deposit_box_erc721(self) -> DepositBoxERC721:
        return DepositBoxERC721(self, MainnetImaContract.DEPOSIT_BOX_ERC721)

    @cached_property
    def deposit_box_erc721_wmt(self) -> DepositBoxERC721WithMetadata:
        return DepositBoxERC721WithMetadata(self, MainnetImaContract.DEPOSIT_BOX_ERC721_WMT)

    @cached_property
    def deposit_box_erc1155(self) -> DepositBoxERC1155:
        return DepositBoxERC1155(self, MainnetImaContract.DEPOSIT_BOX_ERC1155)


def spawn_skale_ima_lib(skale_ima: SkaleIma) -> SkaleIma:
    """Clone skale ima object with the same wallet"""
    return SkaleIma(skale_ima._endpoint, skale_ima.instance.address, skale_ima.wallet)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from functools import cached_property

from skale.skale_base import SkaleBase
from skale_contracts.projects.fair_manager import FairManagerContract
from skale_contracts.project_factory import SkaleProject

from skale.contracts.fair.nodes import Nodes
from skale.contracts.fair.committee import Committee
from skale.contracts.fair.dkg import DKG
from skale.contracts.fair.access_manager import AccessManager
from skale.contracts.fair.status import Status
from skale.contracts.fair.staking import Staking


class FairManager(SkaleBase):
    @property
    def project_name(self) -> SkaleProject:
        return SkaleProject.FAIR_MANAGER

    @cached_property
    def nodes(self) -> Nodes:
        return Nodes(self, FairManagerContract.NODES)

    @cached_property
    def committee(self) -> Committee:
        return Committee(self, FairManagerContract.COMMITTEE)

    @cached_property
    def dkg(self) -> DKG:
        return DKG(self, FairManagerContract.DKG)

    @cached_property
    def access_manager(self) -> AccessManager:
        return AccessManager(self, FairManagerContract.FAIR_ACCESS_MANAGER)

    @cached_property
    def status(self) -> Status:
        return Status(self, FairManagerContract.STATUS)

    @cached_property
    def staking(self) -> Staking:
        return Staking(self, FairManagerContract.STAKING)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import logging

from skale import FairManager
from skale.types.committee import Committee, CommitteeIndex
from skale.types.dkg import G2Point
from skale.types.node import FairNode

logger = logging.getLogger(__name__)

""" This functions are used to generate fair config 'nodeGroups' section data"""


def unpack_bls_public_key(bls_public_key: G2Point) -> dict[str, str]:
    return {
        'blsPublicKey0': str(bls_public_key[0][0]),
        'blsPublicKey1': str(bls_public_key[0][1]),
        'blsPublicKey2': str(bls_public_key[1][0]),
        'blsPublicKey3': str(bls_public_key[1][1]),
    }


def committee_data_to_historical_representation(fair: FairManager, committee: Committee) -> dict:
    bls_public_key = committee.common_public_key
    node_ids = committee.node_ids
    nodes = {}
    for index_in_committee, node_id in enumerate(node_ids):
        node: FairNode = fair.nodes.get(node_id)
        nodes[node.id] = (index_in_committee, node.id, node.public_key)
    committee_data = {
        'rotation': None,
        'nodes': nodes,
        'start_ts': committee.starting_timestamp,
        'bls_public_key': unpack_bls_public_key(bls_public_key),
    }

    return committee_data


def generate_committee_history(fair: FairManager) -> dict:
    latest_committee_index: int = fair.committee.last_committee_index()
    committees = {}

    current_finish_ts = None
    for committee_index in reversed(range(0, latest_committee_index + 1)):
        committee = fair.committee.get_committee(CommitteeIndex(committee_index))
        committee_data = committee_data_to_historical_representation(fair, committee)
        committee_data['finish_ts'] = current_finish_ts
        current_finish_ts = committee_data.pop('start_ts')
        committees.update({str(committee_index): committee_data})
    return committees



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2024-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import NewType

from eth_typing import ChecksumAddress
from web3.types import Wei

from skale.dataclasses.schain_options import SchainOptions


SchainName = NewType('SchainName', str)
SchainHash = NewType('SchainHash', bytes)
SchainOption = tuple[str, bytes]


@dataclass
class Schain:
    name: SchainName
    mainnet_owner: ChecksumAddress
    index_in_owner_list: int
    part_of_node: int
    lifetime: int
    start_date: int
    start_block: int
    deposit: Wei
    index: int
    generation: int
    originator: ChecksumAddress


@dataclass
class SchainStructure(Schain):
    chain_id: SchainHash
    options: SchainOptions


@dataclass
class SchainStructureWithStatus(SchainStructure):
    active: bool



from skale.fair_manager import FairManager
from skale.types.committee import CommitteeGroup, CommitteeIndex
from skale.types.node import FairNode, NodeId

""" This functions are used to generate fair config 'nodes' section data"""


def get_committee_nodes(fair: FairManager, committee_index: int) -> list[FairNode]:
    return [
        fair.nodes.get(NodeId(node_id))
        for node_id in fair.committee.get_committee(CommitteeIndex(committee_index)).node_ids
    ]


def get_nodes_from_last_two_committees(fair: FairManager) -> list[CommitteeGroup]:
    """
    Compose a dictionary with nodes from the last two committees.
    If it is the first committee, it will be saved both
    as first and second committee with first timestamp equal to 0
    """

    latest_committee_index: int = fair.committee.last_committee_index()
    if latest_committee_index == 0:
        committee_a_index: CommitteeIndex = CommitteeIndex(0)
        committee_a = fair.committee.get_committee(CommitteeIndex(0))
        ts_a = 0
    else:
        committee_a_index: CommitteeIndex = CommitteeIndex(latest_committee_index - 1)
        committee_a = fair.committee.get_committee(CommitteeIndex(committee_a_index))
        ts_a = committee_a.starting_timestamp
    committee_a_nodes_data: CommitteeGroup = {
        'index': committee_a_index,
        'ts': ts_a,  # todod: remove, use from committee structure
        'group': get_committee_nodes(fair, committee_a_index),
        'committee': committee_a,
    }

    committee_b_index = latest_committee_index

    committee_b = fair.committee.get_committee(CommitteeIndex(committee_b_index))
    committee_b_nodes_data: CommitteeGroup = {
        'index': committee_b_index,
        'ts': committee_b.starting_timestamp,  # todod: remove, use from committee structure
        'group': get_committee_nodes(fair, committee_b_index),
        'committee': fair.committee.get_committee(committee_b_index),
    }

    committee_nodes_data = [committee_a_nodes_data, committee_b_nodes_data]

    return committee_nodes_data



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from typing import List
from eth_typing import ChecksumAddress

from skale.contracts.base_contract import BaseContract, transaction_method
from skale.types.node import NodeId


class Staking(BaseContract):
    def get_node_share(self, node: NodeId) -> int:
        return self.contract.functions.getNodeShare(node).call()

    def get_staked_amount(self) -> int:
        return self.contract.functions.getStakedAmount().call()

    def get_staked_to_node_amount(self, node: NodeId) -> int:
        return self.contract.functions.getStakedToNodeAmount(node).call(
            {'from': self.skale.wallet.address}
        )

    def get_staked_nodes(self) -> List[NodeId]:
        return self.contract.functions.getStakedNodes().call()

    def get_earned_fee_amount(self, node: NodeId) -> int:
        return self.contract.functions.getEarnedFeeAmount(node).call()

    def get_staked_amount_for(self, holder: ChecksumAddress) -> int:
        return self.contract.functions.getStakedAmountFor(holder).call()

    def get_staked_nodes_for(self, holder: ChecksumAddress) -> List[NodeId]:
        return self.contract.functions.getStakedNodesFor(holder).call()

    def get_staked_to_node_amount_for(self, node: NodeId, holder: ChecksumAddress) -> int:
        return self.contract.functions.getStakedToNodeAmountFor(node, holder).call()

    @transaction_method
    def stake(self, node: NodeId):
        return self.contract.functions.stake(node)

    @transaction_method
    def retrieve(self, node: NodeId, value: int):
        return self.contract.functions.retrieve(node, value)

    @transaction_method
    def set_fee_rate(self, fee_rate: int):
        return self.contract.functions.setFeeRate(fee_rate)

    @transaction_method
    def claim_fee(self, to: ChecksumAddress, amount: int):
        return self.contract.functions.claimFee(to, amount)

    @transaction_method
    def claim_all_fee(self, to: ChecksumAddress):
        return self.contract.functions.claimAllFee(to)

    @transaction_method
    def disable(self, node: NodeId):
        return self.contract.functions.disable(node)

    @transaction_method
    def enable(self, node: NodeId):
        return self.contract.functions.enable(node)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2020-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from web3.contract.contract import ContractFunction

from skale.contracts.base_contract import transaction_method
from skale.contracts.skale_manager_contract import SkaleManagerContract


class TimeHelpersWithDebug(SkaleManagerContract):
    """Wrapper for TimeHelpersWithDebug.sol functions (internal usage only)"""

    @transaction_method
    def skip_time(self, sec: int) -> 'ContractFunction':
        """Skip time on contracts

        :param sec: Time to skip in seconds
        :type sec: int
        :returns: Transaction results
        :rtype: TxRes
        """
        return self.contract.functions.skipTime(sec)

    def get_current_month(self) -> int:
        """Get current month from contract

        :returns: Month index
        :rtype: int
        """
        return int(self.contract.functions.getCurrentMonth().call())



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from typing import List
from eth_typing import ChecksumAddress
from web3.contract.contract import ContractFunction

from skale.contracts.base_contract import transaction_method
from skale.transactions.result import TxRes
from skale.contracts.ima_contract import ImaContract
from skale.types.schain import SchainName


class Linker(ImaContract):
    @transaction_method
    def connect_schain(
        self, schain_name: SchainName, mainnet_contracts: List[ChecksumAddress]
    ) -> 'ContractFunction':
        return self.contract.functions.connectSchain(schain_name, mainnet_contracts)

    def linker_role(self) -> bytes:
        return self.contract.functions.LINKER_ROLE().call()

    def has_role(self, role: bytes, address: ChecksumAddress) -> bool:
        return self.contract.functions.hasRole(role, address).call()

    @transaction_method
    def grant_role(self, role: bytes, address: ChecksumAddress) -> TxRes:
        return self.contract.functions.grantRole(role, address)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os

from eth_typing import HexStr

from skale import SkaleManager
from skale.types.schain import SchainName
from skale.utils.account_tools import send_eth
from skale.utils.contracts_provision.main import (
    DEFAULT_DOMAIN_NAME,
    add_test2_schain_type,
    add_test4_schain_type,
    add_test_permissions,
    create_schain,
    setup_validator,
)
from skale.utils.contracts_provision.utils import generate_random_node_data
from skale.utils.web3_utils import init_web3
from skale.wallets import Web3Wallet
from skale.wallets.web3_wallet import generate_wallets

logger = logging.getLogger(__name__)

TEST_ETH_AMOUNT = 1
NODES_IN_SCHAIN = 4
CHAIN_NAME = SchainName('fair')


def transfer_eth_to_wallets(skale, wallets):
    logger.info(f'Transfering {TEST_ETH_AMOUNT} ETH to {len(wallets)} test wallets')
    for wallet in wallets:
        send_eth(skale.web3, skale.wallet, wallet.address, TEST_ETH_AMOUNT)


def link_addresses_to_validator(skale: SkaleManager, wallets: list[Web3Wallet]) -> None:
    logger.info('Linking addresses to validator')
    for wallet in wallets:
        link_node_address(skale, wallet)


def link_node_address(skale: SkaleManager, wallet: Web3Wallet) -> None:
    validator_id = skale.validator_service.validator_id_by_address(skale.wallet.address)
    main_wallet = skale.wallet
    skale.wallet = wallet
    signature = skale.validator_service.get_link_node_signature(validator_id=validator_id)
    skale.wallet = main_wallet
    skale.validator_service.link_node_address(
        node_address=wallet.address, signature=signature, wait_for=True
    )


def init_skale_from_wallet(skale: SkaleManager, wallet: Web3Wallet) -> SkaleManager:
    return SkaleManager(skale._endpoint, skale._alias_or_address, wallet)


def register_nodes(skale_instances):
    nodes = [register_node(sk) for sk in skale_instances]
    return nodes


def register_node(skale):
    ip, public_ip, port, name = generate_random_node_data()
    port = 10000
    skale.manager.create_node(
        ip=ip,
        port=port,
        name=name,
        public_ip=public_ip,
        domain_name=DEFAULT_DOMAIN_NAME,
        wait_for=True,
    )
    node_id = skale.nodes.node_name_to_index(name)
    logger.info(f'Registered node {name}, ID: {node_id}')
    return {'node': skale.nodes.get_by_name(name), 'node_id': node_id, 'wallet': skale.wallet}


def set_up_nodes(skale, nodes_number, no_zero_id=True):
    if no_zero_id:
        nodes_number += 1
    wallets = generate_wallets(skale.web3, nodes_number)
    transfer_eth_to_wallets(skale, wallets)
    link_addresses_to_validator(skale, wallets)
    skale_instances = [init_skale_from_wallet(skale, wallet) for wallet in wallets]
    nodes_data = register_nodes(skale_instances)
    if no_zero_id:
        skale.nodes.init_exit(0)
        skale.manager.node_exit(0)
        return nodes_data[1:], skale_instances[1:]
    else:
        return nodes_data, skale_instances


def init_skale_manager(
    endpoint: str, alias_or_address: str, eth_private_key: HexStr
) -> SkaleManager:
    web3 = init_web3(endpoint)
    wallet = Web3Wallet(eth_private_key, web3)
    return SkaleManager(endpoint, alias_or_address, wallet)


def bootstrap_fair(endpoint: str, alias_or_address: str, eth_private_key: HexStr) -> None:
    skale = init_skale_manager(endpoint, alias_or_address, eth_private_key)
    add_test_permissions(skale)
    setup_validator(skale)
    add_test2_schain_type(skale)
    add_test4_schain_type(skale)

    set_up_nodes(skale, NODES_IN_SCHAIN)
    create_schain(
        skale,
        schain_name=CHAIN_NAME,
        schain_type=2,
    )


if __name__ == '__main__':
    ENDPOINT = os.environ['ENDPOINT']
    ETH_PRIVATE_KEY = os.environ['ETH_PRIVATE_KEY']
    MANAGER_CONTRACTS = os.environ['MANAGER_CONTRACTS']
    bootstrap_fair(ENDPOINT, MANAGER_CONTRACTS, HexStr(ETH_PRIVATE_KEY))



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2024-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from collections import namedtuple
from typing import List, NamedTuple, NewType, Tuple

from eth_typing import HexStr
from enum import Enum

from skale.types.node import NodeId


Fp2Point = namedtuple('Fp2Point', ['a', 'b'])
DkgId = NewType('DkgId', int)


class G2Point(NamedTuple):
    x: Fp2Point
    y: Fp2Point


VerificationVector = NewType('VerificationVector', List[G2Point])


class KeyShare(NamedTuple):
    publicKey: Tuple[bytes | HexStr, bytes | HexStr]
    share: bytes | HexStr


class Status(Enum):
    SUCCESS = 0
    BROADCAST = 1
    ALRIGHT = 2
    FAILED = 3


class Round(NamedTuple):
    id: DkgId
    status: Status
    nodes: List[NodeId]
    publicKey: G2Point
    startingBlockNumber: int
    numberOfBroadcasted: int
    hashedData: List[bytes]
    numberOfCompleted: int
    completed: List[bool]



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

# flake8: noqa
CONSTELLATIONS = [
    'Andromeda',
    'Antlia',
    'Apus',
    'Aquarius',
    'Aquila',
    'Ara',
    'Aries',
    'Auriga',
    'Caelum',
    'Camelopardalis',
    'Cancer',
    'Canes Venatici',
    'Canis Major',
    'Canis Minor',
    'Capricornus',
    'Carina',
    'Cassiopeia',
    'Centaurus',
    'Cepheus',
    'Cetus',
    'Chamaeleon',
    'Circinus',
    'Columba',
    'Coma Berenices',
    'Corona Austrina',
    'Corona Borealis',
    'Corvus',
    'Crater',
    'Crux',
    'Cygnus',
    'Delphinus',
    'Dorado',
    'Draco',
    'Equuleus',
    'Eridanus',
    'Fornax',
    'Gemini',
    'Grus',
    'Hercules',
    'Horologium',
    'Hydra',
    'Hydrus',
    'Indus',
    'Lacerta',
    'Leo',
    'Leo Minor',
    'Lepus',
    'Libra',
    'Lupus',
    'Lynx',
    'Lyra',
    'Mensa',
    'Microscopium',
    'Monoceros',
    'Musca',
    'Norma',
    'Octans',
    'Ophiuchus',
    'Orion',
    'Pavo',
    'Pegasus',
    'Perseus',
    'Phoenix',
    'Pictor',
    'Pisces',
    'Piscis Austrinus',
    'Puppis',
    'Pyxis',
    'Reticulum',
    'Sagitta',
    'Sagittarius',
    'Scorpius',
    'Sculptor',
    'Scutum',
    'Serpens',
    'Sextans',
    'Taurus',
    'Telescopium',
    'Triangulum',
    'Triangulum Australe',
    'Tucana',
    'Ursa Major',
    'Ursa Minor',
    'Vela',
    'Virgo',
    'Volans',
    'Vulpecula',
]
STARS = [
    'ACAMAR',
    'ACHERNAR',
    'Achird',
    'ACRUX',
    'Acubens',
    'ADARA',
    'Adhafera',
    'Adhil',
    'AGENA',
    'Ain al Rami',
    'Ain',
    'Al Anz',
    'Aladfar',
    'Alathfar',
    'Albaldah',
    'Albali',
    'ALBIREO',
    'Alchiba',
    'ALCOR',
    'ALCYONE',
    'ALDEBARAN',
    'ALDERAMIN',
    'Aldhibah',
    'Alfecca Meridiana',
    'Alfirk',
    'ALGENIB',
    'ALGIEBA',
    'ALGOL',
    'Algorab',
    'ALHENA',
    'ALIOTH',
    'ALKAID',
    'Alkalurops',
    'Alkes',
    'Alkurhah',
    'ALMAAK',
    'ALNAIR',
    'ALNATH',
    'ALNILAM',
    'ALNITAK',
    'Alniyat',
    'Alniyat',
    'ALPHARD',
    'ALPHEKKA',
    'ALPHERATZ',
    'Alrai',
    'Alrisha',
    'Alsafi',
    'Alsciaukat',
    'ALSHAIN',
    'Alshat',
    'Alsuhail',
    'ALTAIR',
    'Altarf',
    'Alterf',
    'Aludra',
    'Alula Australis',
    'Alula Borealis',
    'Alya',
    'Alzirr',
    'Ancha',
    'Angetenar',
    'ANKAA',
    'Anser',
    'ANTARES',
    'ARCTURUS',
    'Arkab Posterior',
    'Arkab Prior',
    'ARNEB',
    'Arrakis',
    'Ascella',
    'Asellus Australis',
    'Asellus Borealis',
    'Asellus Primus',
    'Asellus Secondus',
    'Asellus Tertius',
    'Asterope',
    'Atik',
    'Atlas',
    'Auva',
    'Avior',
    'Azelfafage',
    'Azha',
    'Azmidiske',
    'Baham',
    'Baten Kaitos',
    'Becrux',
    'Beid',
    'BELLATRIX',
    'BETELGEUSE',
    'Botein',
    'Brachium',
    'CANOPUS',
    'CAPELLA',
    'Caph',
    'CASTOR',
    'Cebalrai',
    'Celaeno',
    'Chara',
    'Chort',
    'COR CAROLI',
    'Cursa',
    'Dabih',
    'Deneb Algedi',
    'Deneb Dulfim',
    'Deneb el Okab',
    'DENEB',
    'DENEBOLA',
    'Dheneb',
    'Diadem',
    'DIPHDA',
    'Double Double',
    'Dschubba',
    'Dsiban',
    'DUBHE',
    'Ed Asich',
    'Electra',
    'ELNATH',
    'ENIF',
    'ETAMIN',
    'FOMALHAUT',
    'Fornacis',
    'Fum al Samakah',
    'Furud',
    'Gacrux',
    'Gianfar',
    'Gienah Cygni',
    'Gienah Ghurab',
    'Gomeisa',
    'Gorgonea Quarta',
    'Gorgonea Secunda',
    'Gorgonea Tertia',
    'Graffias',
    'Grafias',
    'Grumium',
    'HADAR',
    'Haedi',
    'HAMAL',
    'Hassaleh',
    'Head of Hydrus',
    'Heze',
    'Hoedus',
    'Homam',
    'Hyadum',
    'IZAR',
    'Jabbah',
    'Kaffaljidhma',
    'Kajam',
    'KAUS AUSTRALIS',
    'Kaus Borealis',
    'Kaus Meridionalis',
    'Keid',
    'Kitalpha',
    'KOCAB',
    'Kornephoros',
    'Kraz',
    'Kuma',
    'Lesath',
    'Maasym',
    'Maia',
    'Marfak',
    'Marfak',
    'Marfic',
    'Marfik',
    'MARKAB',
    'Matar',
    'Mebsuta',
    'MEGREZ',
    'Meissa',
    'Mekbuda',
    'Menkalinan',
    'MENKAR',
    'Menkar',
    'Menkent',
    'Menkib',
    'MERAK',
    'Merga',
    'Merope',
    'Mesarthim',
    'Metallah',
    'Miaplacidus',
    'Minkar',
    'MINTAKA',
    'MIRA',
    'MIRACH',
    'Miram',
    'MIRPHAK',
    'MIZAR',
    'Mufrid',
    'Muliphen',
    'Murzim',
    'Muscida',
    'Muscida',
    'Muscida',
    'Naos',
    'Nash',
    'Nashira',
    'Nekkar',
    'NIHAL',
    'Nodus Secundus',
    'NUNKI',
    'Nusakan',
    'Peacock',
    'PHAD',
    'Phaet',
    'Pherkad Minor',
    'Pherkad',
    'Pleione',
    'Polaris Australis',
    'POLARIS',
    'POLLUX',
    'Porrima',
    'Praecipua',
    'Prima Giedi',
    'PROCYON',
    'Propus',
    'Propus',
    'Propus',
    'Rana',
    'RASALGETHI',
    'RASALHAGUE',
    'Rastaban',
    'REGULUS',
    'Rigel Kentaurus',
    'RIGEL',
    'Rijl al Awwa',
    'Rotanev',
    'Ruchba',
    'Ruchbah',
    'Rukbat',
    'Sabik',
    'Sadalachbia',
    'SADALMELIK',
    'Sadalsuud',
    'Sadr',
    'SAIPH',
    'Salm',
    'Sargas',
    'Sarin',
    'Sceptrum',
    'SCHEAT',
    'Secunda Giedi',
    'Segin',
    'Seginus',
    'Sham',
    'Sharatan',
    'SHAULA',
    'SHEDIR',
    'Sheliak',
    'SIRIUS',
    'Situla',
    'Skat',
    'SPICA',
    'Sterope',
    'Sualocin',
    'Subra',
    'Sulafat',
    'Syrma',
    'Tabit',
    'Talitha',
    'Tania Australis',
    'Tania Borealis',
    'TARAZED',
    'Taygeta',
    'Tegmen',
    'Tejat Posterior',
    'Terebellum',
    'Thabit',
    'Theemim',
    'THUBAN',
    'Turais',
    'Tyl',
    'UNUKALHAI',
    'VEGA',
    'VINDEMIATRIX',
    'Wasat',
    'Wezen',
    'Wezn',
    'Yed Posterior',
    'Yed Prior',
    'Yildun',
    'Zaniah',
    'Zaurak',
    'Zavijah',
    'Zibal',
    'Zosma',
    'Zuben Elakrab',
]

POSITIVE_AND_TIME_ADJECTIVES = [
    'aggressive',
    'agreeable',
    'ambitious',
    'brave',
    'calm',
    'delightful',
    'eager',
    'faithful',
    'gentle',
    'happy',
    'jolly',
    'kind',
    'lively',
    'nice',
    'obedient',
    'polite',
    'proud',
    'silly',
    'thankful',
    'victorious',
    'witty',
    'wonderful',
    'zealous',
    'ancient',
    'brief',
    'early',
    'fast',
    'future',
    'late',
    'long',
    'modern',
    'old',
    'old-fashioned',
    'prehistoric',
    'quick',
    'rapid',
    'short',
    'slow',
    'swift',
    'young',
]
SOUND_AND_APPEARANCE_ADJECTIVES = [
    'deafening',
    'echoing',
    'faint',
    'harsh',
    'hissing',
    'howling',
    'loud',
    'melodic',
    'noisy',
    'purring',
    'quiet',
    'rapping',
    'raspy',
    'rhythmic',
    'screeching',
    'shrilling',
    'squeaking',
    'thundering',
    'tinkling',
    'wailing',
    'whining',
    'whispering',
    'attractive',
    'beautiful',
    'chubby',
    'clean',
    'dazzling',
    'drab',
    'elegant',
    'fancy',
    'fit',
    'glamorous',
    'gorgeous',
    'handsome',
    'long',
    'magnificent',
    'plain',
    'plump',
    'quaint',
    'scruffy',
    'shapely',
    'stocky',
]



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2020-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from typing import List
from skale.types.dkg import G2Point
from skale.contracts.skale_manager_contract import SkaleManagerContract
from skale.types.schain import SchainHash


class KeyStorage(SkaleManagerContract):
    def get_common_public_key(self, schain_hash: SchainHash) -> G2Point:
        return G2Point(*self.contract.functions.getCommonPublicKey(schain_hash).call())

    def get_previous_public_key(self, schain_hash: SchainHash) -> G2Point:
        return G2Point(*self.contract.functions.getPreviousPublicKey(schain_hash).call())

    def get_all_previous_public_keys(self, schain_hash: SchainHash) -> List[G2Point]:
        return [
            G2Point(*key)
            for key in self.contract.functions.getAllPreviousPublicKeys(schain_hash).call()
        ]



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.


import logging
import time
from functools import partial, wraps
from typing import Any, Callable, Optional, TYPE_CHECKING

from eth_typing import ChecksumAddress
from web3 import Web3
from web3.contract.contract import ContractFunction
from web3.exceptions import ContractLogicError, Web3Exception
from web3._utils.transactions import get_block_gas_limit
from web3.types import Nonce, TxParams, Wei

import skale.config as config
from skale.transactions.exceptions import TransactionError
from skale.transactions.result import TxCallResult, TxRes, TxStatus
from skale.utils.web3_utils import get_eth_nonce

if TYPE_CHECKING:
    from skale.skale_base import SkaleBase


logger = logging.getLogger(__name__)


DEFAULT_ETH_SEND_GAS_LIMIT = 22000


def make_dry_run_call(
    skale: 'SkaleBase', method: ContractFunction, gas_limit: int | None = None, value: Wei = Wei(0)
) -> TxCallResult:
    opts = TxParams({'from': skale.wallet.address, 'value': value})
    logger.info(
        f'Dry run tx: {method.fn_name}, '
        f'sender: {skale.wallet.address}, '
        f'wallet: {skale.wallet.__class__.__name__}, '
        f'value: {value}, '
    )
    estimated_gas = 0

    try:
        if gas_limit:
            estimated_gas = gas_limit
            opts.update({'gas': gas_limit})
            method.call(opts)
        else:
            estimated_gas = estimate_gas(skale.web3, method, opts)
        logger.info(f'Estimated gas for {method.fn_name}: {estimated_gas}')
    except ContractLogicError as e:
        message = e.message or 'Contract logic error'
        error_data = e.data or {}
        data = {'data': error_data} if isinstance(error_data, str) else error_data
        return TxCallResult(status=TxStatus.FAILED, error='revert', message=message, data=data)
    except (Web3Exception, ValueError) as e:
        logger.exception('Dry run for %s failed', method)
        return TxCallResult(status=TxStatus.FAILED, error='exception', message=str(e), data={})

    return TxCallResult(
        status=TxStatus.SUCCESS, error='', message='success', data={'gas': estimated_gas}
    )


def estimate_gas(web3: Web3, method: ContractFunction, opts: TxParams) -> int:
    try:
        block_gas_limit = get_block_gas_limit(web3)
    except AttributeError:
        block_gas_limit = get_block_gas_limit(web3)

    estimated_gas = method.estimate_gas(opts, block_identifier='latest')
    normalized_estimated_gas = int(estimated_gas * config.DEFAULT_GAS_MULTIPLIER)
    if normalized_estimated_gas > block_gas_limit:
        logger.warning(
            f'Estimate gas for {method.fn_name} - {normalized_estimated_gas} exceeds \
block gas limit, going to use block_gas_limit ({block_gas_limit}) for this transaction'
        )
        return block_gas_limit
    return normalized_estimated_gas


def build_tx_dict(method: ContractFunction, *args: Any, **kwargs: Any) -> TxParams:
    base_fields = compose_base_fields(*args, **kwargs)
    return method.build_transaction(base_fields)


def compose_base_fields(
    nonce: Nonce,
    gas_limit: int,
    gas_price: Wei | None = None,
    max_fee_per_gas: Wei | None = None,
    max_priority_fee_per_gas: Wei | None = None,
    value: Wei = Wei(0),
) -> TxParams:
    fee_fields = TxParams({'gas': gas_limit, 'nonce': nonce, 'value': value})
    if max_priority_fee_per_gas is not None and max_fee_per_gas is not None:
        fee_fields.update(
            {'maxPriorityFeePerGas': max_priority_fee_per_gas, 'maxFeePerGas': max_fee_per_gas}
        )
        fee_fields.update({'type': 2})
    elif gas_price is not None:
        fee_fields.update({'gasPrice': gas_price})
        fee_fields.update({'type': 1})
    return fee_fields


def transaction_from_method(
    method: ContractFunction,
    *,
    multiplier: Optional[float] = None,
    priority: Optional[int] = None,
    **kwargs: Any,
) -> TxParams:
    tx = build_tx_dict(method, **kwargs)
    logger.info(f'Tx: {method.fn_name}, Fields: {tx}, ')
    return tx


def compose_eth_transfer_tx(
    web3: Web3,
    from_address: ChecksumAddress,
    to_address: ChecksumAddress,
    value: Wei,
    **kwargs: Any,
) -> TxParams:
    nonce = get_eth_nonce(web3, from_address)
    base_fields = compose_base_fields(
        nonce=nonce, gas_limit=DEFAULT_ETH_SEND_GAS_LIMIT, value=value, **kwargs
    )
    tx = TxParams({'from': from_address, 'to': to_address, **base_fields})
    return tx


def retry_tx(
    tx: Callable[..., TxRes] | None = None, *, max_retries: int = 3, timeout: int = -1
) -> Callable[..., TxRes | None] | partial[Any]:
    if tx is None:
        return partial(retry_tx, max_retries=3, timeout=timeout)

    @wraps(tx)
    def wrapper(*args: Any, **kwargs: Any) -> TxRes | None:
        return run_tx_with_retry(
            tx, *args, max_retries=max_retries, retry_timeout=timeout, **kwargs
        )

    return wrapper


def run_tx_with_retry(
    transaction: Callable[..., TxRes],
    *args: Any,
    max_retries: int = 3,
    retry_timeout: int = -1,
    raise_for_status: bool = True,
    **kwargs: Any,
) -> TxRes | None:
    attempt = 0
    tx_res = None
    exp_timeout = 1
    error = None
    while attempt < max_retries:
        try:
            tx_res = transaction(*args, raise_for_status=raise_for_status, **kwargs)
            tx_res.raise_for_status()
        except TransactionError as e:
            error = e
            logger.exception('Tx attempt %d/%d failed', attempt + 1, max_retries)

            timeout = exp_timeout if retry_timeout < 0 else exp_timeout
            time.sleep(timeout)
            exp_timeout *= 2
        else:
            error = None
            break
        attempt += 1
    if error is None:
        logger.info(
            'Tx %s completed after %d/%d retries', transaction.__name__, attempt + 1, max_retries
        )
    else:
        logger.error('Tx %s failed after %d retries', transaction.__name__, max_retries)
        if raise_for_status:
            raise error
    if tx_res is not None:
        tx_res.attempts = attempt
    return tx_res



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from skale.contracts.base_contract import BaseContract, transaction_method
from skale.transactions.result import TxRes
from Crypto.Hash import keccak
from eth_typing import ChecksumAddress
from skale.types.schain import SchainName


class DepositBoxERC20(BaseContract):
    """Deposit Box"""

    def is_whitelisted(self, schain_name: SchainName) -> bool:
        return self.contract.functions.isWhitelisted(schain_name).call()

    @transaction_method
    def enable_whitelist(self, schain_name: SchainName) -> TxRes:
        return self.contract.functions.enableWhitelist(schain_name)

    @transaction_method
    def disable_whitelist(self, schain_name: SchainName) -> TxRes:
        return self.contract.functions.disableWhitelist(schain_name)

    @transaction_method
    def add_erc20_token(self, schain_name: SchainName, address: ChecksumAddress) -> TxRes:
        return self.contract.functions.addERC20TokenByOwner(schain_name, address)

    @transaction_method
    def deposit_erc20(
        self, schain_name: SchainName, address: ChecksumAddress, amount: int
    ) -> TxRes:
        return self.contract.functions.depositERC20(schain_name, address, amount)

    @transaction_method
    def deposit_erc20_direct(
        self, schain_name: SchainName, address: ChecksumAddress, amount: int, receiver: int
    ) -> TxRes:
        return self.contract.functions.depositERC20Direct(schain_name, address, amount, receiver)

    @transaction_method
    def set_big_transfer_value(self, schain_name: SchainName, token: int, value: int) -> TxRes:
        return self.contract.functions.setBigTransferValue(schain_name, token, value)

    @transaction_method
    def set_big_transfer_delay(self, schain_name: SchainName, delay: int) -> TxRes:
        return self.contract.functions.setBigTransferDelay(schain_name, delay)

    @transaction_method
    def set_arbitrage_duration(self, schain_name: SchainName, delay: int) -> TxRes:
        return self.contract.functions.setArbitrageDuration(schain_name, delay)

    @transaction_method
    def trust_receiver(self, schain_name: SchainName, address: ChecksumAddress) -> TxRes:
        return self.contract.functions.trustReceiver(schain_name, address)

    def is_receiver_trusted(self, schain_name: SchainName, address: ChecksumAddress) -> bool:
        keccak_hash = keccak.new(data=schain_name.encode('utf8'), digest_bits=256)
        schain_id = keccak_hash.digest()
        return self.contract.functions.isReceiverTrusted(schain_id, address).call()

    def arbiter_role(self) -> bytes:
        return self.contract.functions.ARBITER_ROLE().call()

    def admin_role(self) -> bytes:
        return self.contract.functions.DEFAULT_ADMIN_ROLE().call()

    def has_role(self, role: bytes, address: ChecksumAddress) -> bool:
        return self.contract.functions.hasRole(role, address).call()

    @transaction_method
    def grant_role(self, role: bytes, address: ChecksumAddress) -> TxRes:
        return self.contract.functions.grantRole(role, address)

    def get_role_member(self, role: bytes, index: int) -> bytes:
        return self.contract.functions.getRoleMember(role, index).call()



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2024-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from dataclasses import dataclass
from typing import List, NewType, TypedDict

from skale.types.dkg import DkgId, G2Point
from skale.types.node import FairNode, NodeId

CommitteeIndex = NewType('CommitteeIndex', int)
Timestamp = NewType('Timestamp', int)


@dataclass
class Committee:
    node_ids: List[NodeId]
    dkg_id: DkgId
    common_public_key: G2Point
    starting_timestamp: Timestamp


CommitteeGroup = TypedDict(
    'CommitteeGroup',
    {
        'ts': Timestamp,
        'index': CommitteeIndex,
        'group': List[FairNode],
        'committee': Committee,
    },
)



from skale.wallets.common import BaseWallet
from skale.wallets.ledger_wallet import LedgerWallet
from skale.wallets.redis_wallet import RedisWalletAdapter
from skale.wallets.sgx_wallet import SgxWallet
from skale.wallets.web3_wallet import Web3Wallet

__all__ = ['BaseWallet', 'LedgerWallet', 'Web3Wallet', 'SgxWallet', 'RedisWalletAdapter']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import (
    find_packages,
    setup,
)

extras_require = {
    'linter': [
        'ruff==0.12.2',
        'isort>=4.2.15,<5.4.3',
        'importlib-metadata<5.0',
    ],
    'dev': [
        'bumpversion==0.6.0',
        'click==7.1.2',
        'freezegun==1.2.2',
        'mock==4.0.2',
        'pytest==7.1.3',
        'pytest-cov==2.8.1',
        'Random-Word==1.0.4',
        'twine==4.0.2',
        'when-changed',
    ],
    'hw-wallet': ['ledgerblue==0.1.47'],
}

extras_require['dev'] = (
    extras_require['linter'] + extras_require['dev'] + extras_require['hw-wallet']
)


setup(
    name='skale.py',
    version='7.3dev2',
    description='SKALE client tools',
    long_description_markdown_filename='README.md',
    author='SKALE Labs',
    author_email='support@skalelabs.com',
    url='https://github.com/skalenetwork/skale.py',
    include_package_data=True,
    install_requires=[
        'pyyaml==6.0',
        'redis==6.2.0',
        'sgx.py==0.10dev0',
        'skale-contracts==2.0.0a4',
        'typing-extensions==4.14.1',
        'web3==7.12.0',
    ],
    python_requires='>=3.11,<4',
    extras_require=extras_require,
    keywords='skale',
    packages=find_packages(exclude=['tests']),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.11',
    ],
)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from eth_typing import ChecksumAddress
from skale.contracts.base_contract import BaseContract, transaction_method
from skale.transactions.result import TxRes
from Crypto.Hash import keccak
from skale.types.schain import SchainName


class CommunityLocker(BaseContract):
    """ "Community locker"""

    @transaction_method
    def set_time_limit_per_message(self, schain_name: SchainName, new_time_limit: int) -> TxRes:
        """Set time limit"""
        return self.contract.functions.setTimeLimitPerMessage(schain_name, new_time_limit)

    def constant_setter_role(self) -> bytes:
        return self.contract.functions.CONSTANT_SETTER_ROLE().call()

    def has_role(self, role: bytes, address: ChecksumAddress) -> bool:
        return self.contract.functions.hasRole(role, address).call()

    @transaction_method
    def revoke_role(self, role: bytes, address: ChecksumAddress) -> TxRes:
        return self.contract.functions.revokeRole(role, address)

    @transaction_method
    def grant_role(self, role: bytes, address: ChecksumAddress) -> TxRes:
        return self.contract.functions.grantRole(role, address)

    def check_allow_to_send_msg(self, schain_name: SchainName, address: ChecksumAddress) -> int:
        keccak_hash = keccak.new(data=schain_name.encode('utf8'), digest_bits=256)
        hash = keccak_hash.digest()
        return self.contract.functions.checkAllowedToSendMessage(hash, address).call()

    def schain_hash(self) -> bytes:
        return self.contract.functions.schainHash().call()

    def mainnet_hash(self) -> bytes:
        return self.contract.functions.MAINNET_HASH().call()

    def is_active_user(self, address) -> bool:
        return self.contract.functions.activeUsers(address).call()

    def last_message_timestamp(self, address) -> int:
        return self.contract.functions.lastMessageTimeStamp(address).call()

    def time_limit_per_msg(self, schain_name: SchainName) -> int:
        keccak_hash = keccak.new(data=schain_name.encode('utf8'), digest_bits=256)
        hash = keccak_hash.digest()
        return self.contract.functions.timeLimitPerMessage(hash).call()



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2024-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from enum import IntEnum
from typing import NewType, TypedDict

from web3.types import Wei


class TimeUnit(IntEnum):
    DAY = 0
    MONTH = 1
    YEAR = 2


class BeneficiaryStatus(IntEnum):
    UNKNOWN = 0
    CONFIRMED = 1
    ACTIVE = 2
    TERMINATED = 3


PlanId = NewType('PlanId', int)


class Plan(TypedDict):
    totalVestingDuration: int
    vestingCliff: int
    vestingIntervalTimeUnit: TimeUnit
    vestingInterval: int
    isDelegationAllowed: bool
    isTerminatable: bool


class PlanWithId(Plan):
    planId: PlanId


class BeneficiaryPlan(TypedDict):
    status: BeneficiaryStatus
    statusName: str
    planId: PlanId
    startMonth: int
    fullAmount: Wei
    amountAfterLockup: Wei



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
"""Nodes.sol functions"""

import socket
from typing import Any, Dict, List, Tuple, cast

from Crypto.Hash import keccak
from eth_typing import BlockNumber, ChecksumAddress
from web3.contract.contract import ContractFunction
from web3.exceptions import BadFunctionCallOutput, ContractLogicError

from skale.contracts.base_contract import transaction_method

from skale.contracts.skale_manager_contract import SkaleManagerContract
from skale.types.node import Node, NodeId, NodeStatus, Port
from skale.types.validator import ValidatorId
from skale.utils.exceptions import InvalidNodeIdError
from skale.utils.helper import format_fields

FIELDS = [
    'name',
    'ip',
    'publicIP',
    'port',
    'start_block',
    'last_reward_date',
    'finish_time',
    'status',
    'validator_id',
    'publicKey',
    'domain_name',
]


class Nodes(SkaleManagerContract):
    def __get_raw(self, node_id: NodeId) -> List[Any]:
        try:
            return list(self.contract.functions.nodes(node_id).call())
        except (ContractLogicError, ValueError, BadFunctionCallOutput):
            raise InvalidNodeIdError(node_id)

    def __get_raw_w_pk(self, node_id: NodeId) -> List[Any]:
        raw_node_struct = self.__get_raw(node_id)
        raw_node_struct.append(self.get_node_public_key(node_id))
        return raw_node_struct

    def __get_raw_w_pk_w_domain(self, node_id: NodeId) -> List[Any]:
        raw_node_struct_w_pk = self.__get_raw_w_pk(node_id)
        raw_node_struct_w_pk.append(self.get_domain_name(node_id))
        return raw_node_struct_w_pk

    @format_fields(FIELDS)
    def untyped_get(self, node_id: NodeId) -> List[Any]:
        return self.__get_raw_w_pk_w_domain(node_id)

    def get(self, node_id: NodeId) -> Node:
        node = self.untyped_get(node_id)
        if node is None:
            raise ValueError('Node with id ', node_id, ' is not found')
        if isinstance(node, dict):
            return self._to_node(node)
        if isinstance(node, list):
            return self._to_node(node[0])
        raise ValueError("Can't process returned node type")

    @format_fields(FIELDS)
    def get_by_name(self, name: str) -> List[Any]:
        name_hash = self.name_to_id(name)
        _id = self.contract.functions.nodesNameToIndex(name_hash).call()
        return self.__get_raw_w_pk_w_domain(_id)

    def get_nodes_number(self) -> int:
        return int(self.contract.functions.getNumberOfNodes().call())

    def get_active_node_ids(self) -> List[NodeId]:
        nodes_number = self.get_nodes_number()
        return [
            NodeId(node_id)
            for node_id in range(0, nodes_number)
            if self.get_node_status(NodeId(node_id)) == NodeStatus.ACTIVE
        ]

    def get_active_node_ips(self) -> List[bytes]:
        nodes_number = self.get_nodes_number()
        return [
            self.get(NodeId(node_id))['ip']
            for node_id in range(0, nodes_number)
            if self.get_node_status(NodeId(node_id)) == NodeStatus.ACTIVE
        ]

    def name_to_id(self, name: str) -> bytes:
        keccak_hash = keccak.new(data=name.encode('utf8'), digest_bits=256)
        return keccak_hash.digest()

    def is_node_name_available(self, name: str) -> bool:
        node_id = self.name_to_id(name)
        return not self.contract.functions.nodesNameCheck(node_id).call()

    def is_node_ip_available(self, ip: str) -> bool:
        ip_bytes = socket.inet_aton(ip)
        return not self.contract.functions.nodesIPCheck(ip_bytes).call()

    def node_name_to_index(self, name: str) -> int:
        name_hash = self.name_to_id(name)
        return int(self.contract.functions.nodesNameToIndex(name_hash).call())

    def get_node_status(self, node_id: NodeId) -> NodeStatus:
        try:
            return NodeStatus(self.contract.functions.getNodeStatus(node_id).call())
        except (ContractLogicError, ValueError, BadFunctionCallOutput):
            raise InvalidNodeIdError(node_id)

    def get_node_finish_time(self, node_id: NodeId) -> int:
        try:
            return int(self.contract.functions.getNodeFinishTime(node_id).call())
        except (ContractLogicError, ValueError, BadFunctionCallOutput):
            raise InvalidNodeIdError(node_id)

    def __get_node_public_key_raw(self, node_id: NodeId) -> Tuple[bytes, bytes]:
        try:
            return cast(
                Tuple[bytes, bytes], self.contract.functions.getNodePublicKey(node_id).call()
            )
        except (ContractLogicError, ValueError, BadFunctionCallOutput):
            raise InvalidNodeIdError(node_id)

    def get_node_public_key(self, node_id: NodeId) -> str:
        raw_key = self.__get_node_public_key_raw(node_id)
        key_bytes = raw_key[0] + raw_key[1]
        return self.skale.web3.to_hex(key_bytes)

    def get_validator_node_indices(self, validator_id: int) -> list[NodeId]:
        """Returns list of node indices to the validator

        :returns: List of trusted node indices
        :rtype: list
        """
        return [
            NodeId(id)
            for id in self.contract.functions.getValidatorNodeIndexes(validator_id).call()
        ]

    def get_last_change_ip_time(self, node_id: NodeId) -> int:
        return int(self.contract.functions.getLastChangeIpTime(node_id).call())

    @transaction_method
    def set_node_in_maintenance(self, node_id: NodeId) -> 'ContractFunction':
        return self.contract.functions.setNodeInMaintenance(node_id)

    @transaction_method
    def remove_node_from_in_maintenance(self, node_id: NodeId) -> 'ContractFunction':
        return self.contract.functions.removeNodeFromInMaintenance(node_id)

    @transaction_method
    def set_domain_name(self, node_id: NodeId, domain_name: str) -> 'ContractFunction':
        return self.contract.functions.setDomainName(node_id, domain_name)

    def get_domain_name(self, node_id: NodeId) -> str:
        return str(self.contract.functions.getNodeDomainName(node_id).call())

    @transaction_method
    def grant_role(self, role: bytes, owner: ChecksumAddress) -> 'ContractFunction':
        return self.contract.functions.grantRole(role, owner)

    def has_role(self, role: bytes, address: ChecksumAddress) -> bool:
        return bool(self.contract.functions.hasRole(role, address).call())

    def node_manager_role(self) -> bytes:
        return bytes(self.contract.functions.NODE_MANAGER_ROLE().call())

    def compliance_role(self) -> bytes:
        return bytes(self.contract.functions.COMPLIANCE_ROLE().call())

    @transaction_method
    def init_exit(self, node_id: NodeId) -> 'ContractFunction':
        return self.contract.functions.initExit(node_id)

    @transaction_method
    def change_ip(self, node_id: NodeId, ip: bytes, public_ip: bytes) -> 'ContractFunction':
        return self.contract.functions.changeIP(node_id, ip, public_ip)

    def _to_node(self, untyped_node: Dict[str, Any]) -> Node:
        for key in Node.__annotations__:
            if key not in untyped_node:
                raise ValueError(f'Key: {key} is not available in node.')
        return Node(
            {
                'name': str(untyped_node['name']),
                'ip': bytes(untyped_node['ip']),
                'publicIP': bytes(untyped_node['publicIP']),
                'port': Port(untyped_node['port']),
                'start_block': BlockNumber(untyped_node['start_block']),
                'last_reward_date': int(untyped_node['last_reward_date']),
                'finish_time': int(untyped_node['finish_time']),
                'status': NodeStatus(untyped_node['status']),
                'validator_id': ValidatorId(untyped_node['validator_id']),
                'publicKey': str(untyped_node['publicKey']),
                'domain_name': str(untyped_node['domain_name']),
            }
        )



from eth_typing import ChecksumAddress
from web3.contract.contract import ContractFunction
from web3.types import Wei

from skale.contracts.base_contract import transaction_method
from skale.contracts.skale_manager_contract import SkaleManagerContract


class SlashingTable(SkaleManagerContract):
    """Wrapper for SlashingTable.sol functions"""

    @transaction_method
    def set_penalty(self, offense: str, penalty: Wei) -> 'ContractFunction':
        """Set slashing penalty
        :param offense: reason of slashing
        :type offense: str
        :param penalty: penalty value to set
        :type penalty: int
        :rtype: TxRes
        """
        return self.contract.functions.setPenalty(offense, penalty)

    def get_penalty(self, offense: str) -> Wei:
        """Get slashing penalty value
        :param offense: reason of slashing
        :type offense: str
        :rtype: int
        """
        return Wei(self.contract.functions.getPenalty(offense).call())

    @transaction_method
    def grant_role(self, role: bytes, address: ChecksumAddress) -> 'ContractFunction':
        return self.contract.functions.grantRole(role, address)

    def penalty_setter_role(self) -> bytes:
        return bytes(self.contract.functions.PENALTY_SETTER_ROLE().call())

    def has_role(self, role: bytes, address: ChecksumAddress) -> bool:
        return bool(self.contract.functions.hasRole(role, address).call())



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.


import abc
import logging

from skale_contracts import skale_contracts
from skale_contracts.project_factory import SkaleProject

from skale.utils.exceptions import InvalidWalletError, EmptyWalletError
from skale.utils.web3_utils import default_gas_price, init_web3, get_endpoint
from skale.wallets import BaseWallet


logger = logging.getLogger(__name__)


class EmptyPrivateKey(Exception):
    pass


class SkaleBase:
    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        endpoint: str | list[str],
        alias_or_address: str,
        wallet: BaseWallet | None = None,
        state_path: str | None = None,
        ts_diff: int | None = None,
        provider_timeout: int = 30,
        debug: bool = False,
    ):
        logger.info(
            'Initializing %s, endpoint: %s, alias_or_address: %s, wallet: %s',
            self.__class__.__name__,
            endpoint,
            alias_or_address,
            type(wallet).__name__,
        )
        if state_path:
            logger.warning(
                'state_path is deprecated and will be ignored. This option will be removed in v8.'
            )
        self._endpoint = get_endpoint(endpoint)
        self._alias_or_address = alias_or_address
        self.web3 = init_web3(self._endpoint, ts_diff=ts_diff, provider_timeout=provider_timeout)
        self.network = skale_contracts.get_network_by_provider(self.web3.provider)
        self.project = self.network.get_project(self.project_name)
        self.instance = self.project.get_instance(alias_or_address)
        self.debug = debug
        if wallet:
            self.wallet = wallet

    @property
    @abc.abstractmethod
    def project_name(self) -> SkaleProject:
        """Name of smart contracts project"""

    @property
    def gas_price(self) -> int:
        return default_gas_price(self.web3)

    @property
    def wallet(self) -> BaseWallet:
        if not self._wallet:
            raise EmptyWalletError('No wallet provided')
        return self._wallet

    @wallet.setter
    def wallet(self, wallet: BaseWallet) -> None:
        if issubclass(type(wallet), BaseWallet):
            self._wallet = wallet
        else:
            raise InvalidWalletError(
                f'Wrong wallet class: {type(wallet).__name__}. \
                Must be one of the BaseWallet subclasses'
            )



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any

from skale.contracts.base_contract import BaseContract
from skale.contracts.base_contract import transaction_method
from skale.types.dkg import Fp2Point, G2Point, DkgId, KeyShare, VerificationVector, Status, Round
from skale.types.node import NodeId


class DKG(BaseContract):
    def is_node_broadcasted(self, dkg: DkgId, node: NodeId) -> bool:
        return self.contract.functions.isNodeBroadcasted(dkg, node).call()

    def is_node_sent_alright(self, dkg: DkgId, node: NodeId) -> bool:
        round_info = self.get_round(dkg)
        return round_info.completed[node]

    def get_participants(self, dkg: DkgId) -> list[NodeId]:
        return self.contract.functions.getParticipants(dkg).call()

    def get_public_key(self, dkg: DkgId) -> G2Point:
        return self.contract.functions.getPublicKey(dkg).call()

    def get_last_dkg_id(self) -> DkgId:
        return DkgId(self.contract.functions.lastDkgId().call())

    def get_starting_block_number(self, dkg: DkgId) -> int:
        return self.contract.functions.getStartingBlockNumber(dkg).call()

    def __get_raw_round(self, dkg: DkgId) -> list[Any]:
        return list(self.contract.functions.getRound(dkg).call())

    def get_round(self, dkg: DkgId) -> Round:
        return self._to_round(self.__get_raw_round(dkg))

    def _to_round(self, untyped_round: list[Any]) -> Round:
        return Round(
            id=DkgId(untyped_round[0]),
            status=Status(untyped_round[1]),
            nodes=untyped_round[2],
            publicKey=G2Point(
                x=Fp2Point(a=untyped_round[3][0][0], b=untyped_round[3][0][1]),
                y=Fp2Point(a=untyped_round[3][1][0], b=untyped_round[3][1][1]),
            ),
            startingBlockNumber=untyped_round[4],
            numberOfBroadcasted=untyped_round[5],
            hashedData=untyped_round[6],
            numberOfCompleted=untyped_round[7],
            completed=untyped_round[8],
        )

    def is_last_dkg_successful(self) -> bool:
        last_dkg_id = self.get_last_dkg_id()
        round_info = self.get_round(last_dkg_id)
        return round_info.status == Status.SUCCESS

    @transaction_method
    def alright(self, dkg: DkgId):
        return self.contract.functions.alright(dkg)

    @transaction_method
    def broadcast(
        self,
        dkg: DkgId,
        verification_vector: VerificationVector,
        secret_key_contribution: list[KeyShare],
    ):
        return self.contract.functions.broadcast(dkg, verification_vector, secret_key_contribution)

    @transaction_method
    def generate(self, participants: list[NodeId]):
        return self.contract.functions.generate(participants)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any

from eth_typing import ChecksumAddress

from skale.contracts.base_contract import BaseContract
from skale.contracts.base_contract import transaction_method
from skale.types.committee import Committee as CommitteeStruct, CommitteeIndex
from skale.types.node import NodeId


class Committee(BaseContract):
    def __get_raw(self, committee_index: CommitteeIndex) -> list[Any]:
        return list(self.contract.functions.getCommittee(committee_index).call())

    def get_committee(self, committee_index: CommitteeIndex) -> CommitteeStruct:
        return self._to_committee(self.__get_raw(committee_index))

    def _to_committee(self, untyped_committee: list[Any]) -> CommitteeStruct:
        return CommitteeStruct(
            node_ids=untyped_committee[0],
            dkg_id=untyped_committee[1],
            common_public_key=untyped_committee[2],
            starting_timestamp=untyped_committee[3],
        )

    def is_node_in_current_or_next_committee(self, node: NodeId) -> bool:
        return self.contract.functions.isNodeInCurrentOrNextCommittee(node).call()

    def get_active_committee_index(self) -> CommitteeIndex:
        return CommitteeIndex(self.contract.functions.getActiveCommitteeIndex().call())

    def last_committee_index(self) -> CommitteeIndex:
        return CommitteeIndex(self.contract.functions.lastCommitteeIndex().call())

    def skale_rng(self) -> ChecksumAddress:
        return self.contract.functions.skaleRng().call()

    @transaction_method
    def set_rng(self, address: ChecksumAddress):
        return self.contract.functions.setRNG(address)

    @transaction_method
    def disableRNG(self):
        return self.contract.functions.disableRNG()

    @transaction_method
    def select(self):
        return self.contract.functions.select()

    @transaction_method
    def set_dkg(self, dkg_address: ChecksumAddress):
        return self.contract.functions.setDkg(dkg_address)

    @transaction_method
    def set_nodes(self, nodes_address: ChecksumAddress):
        return self.contract.functions.setNodes(nodes_address)

    @transaction_method
    def set_status(self, status_address: ChecksumAddress):
        return self.contract.functions.setStatus(status_address)

    @transaction_method
    def set_staking(self, staking_address: ChecksumAddress):
        return self.contract.functions.setStaking(staking_address)

    @transaction_method
    def set_version(self, new_version: str):
        return self.contract.functions.setVersion(new_version)

    @transaction_method
    def set_committee_size(self, size: int):
        return self.contract.functions.setCommitteeSize(size)

    @transaction_method
    def set_transition_delay(self, delay: int):
        return self.contract.functions.setTransitionDelay(delay)
